@@ -1,4 +1,7 @@
+# encoding: utf-8
+
 require 'set'
+require 'json'
 require 'puppet-lint/version'
 require 'puppet-lint/lexer'
 require 'puppet-lint/configuration'
@@ -46,7 +49,7 @@ class PuppetLint
   # Public: Initialise a new PuppetLint object.
   def initialize
     @code = nil
-    @statistics = {:error => 0, :warning => 0, :fixed => 0, :ignored => 0}
+    @statistics = { :error => 0, :warning => 0, :fixed => 0, :ignored => 0 }
     @manifest = ''
   end
 
@@ -69,10 +72,16 @@ class PuppetLint
   #
   # Returns nothing.
   def file=(path)
-    if File.exist? path
-      @path = path
-      @code = File.read(path)
+    return unless File.exist?(path)
+
+    @path = path
+    File.open(path, 'r:UTF-8') do |f|
+      @code = f.read
     end
+
+    # Check if the input is an SE Linux policy package file (which also use
+    # the .pp extension), which all have the first 4 bytes 0xf97cff8f.
+    @code = '' if @code[0..3].unpack('V').first == 0xf97cff8f
   end
 
   # Internal: Retrieve the format string to be used when writing problems to
@@ -81,15 +90,14 @@ class PuppetLint
   #
   # Returns a format String to be used with String#%.
   def log_format
-    if configuration.log_format == ''
+    if configuration.log_format.nil? || configuration.log_format.empty?
       ## recreate previous old log format as far as thats possible.
       format = '%{KIND}: %{message} on line %{line}'
-      if configuration.with_filename
-        format.prepend '%{path} - '
-      end
+      format.prepend('%{path} - ') if configuration.with_filename
       configuration.log_format = format
     end
-    return configuration.log_format
+
+    configuration.log_format
   end
 
   # Internal: Format a problem message and print it to STDOUT.
@@ -100,9 +108,17 @@ class PuppetLint
   def format_message(message)
     format = log_format
     puts format % message
-    if message[:kind] == :ignored && !message[:reason].nil?
-      puts "  #{message[:reason]}"
-    end
+
+    puts "  #{message[:reason]}" if message[:kind] == :ignored && !message[:reason].nil?
+  end
+
+  # Internal: Get the line of the manifest on which the problem was found
+  #
+  # message - A Hash containing all the information about a problem.
+  #
+  # Returns the problematic line as a string.
+  def get_context(message)
+    PuppetLint::Data.manifest_lines[message[:line] - 1].strip
   end
 
   # Internal: Print out the line of the manifest on which the problem was found
@@ -114,10 +130,10 @@ class PuppetLint
   def print_context(message)
     return if message[:check] == 'documentation'
     return if message[:kind] == :fixed
-    line = PuppetLint::Data.manifest_lines[message[:line] - 1]
-    offset = line.index(/\S/) || 1
+    line = get_context(message)
+    offset = line.index(%r{\S}) || 1
     puts "\n  #{line.strip}"
-    printf "%#{message[:column] + 2 - offset}s\n\n", '^'
+    printf("%#{message[:column] + 2 - offset}s\n\n", '^')
   end
 
   # Internal: Print the reported problems with a manifest to stdout.
@@ -127,17 +143,25 @@ class PuppetLint
   #
   # Returns nothing.
   def report(problems)
+    json = []
     problems.each do |message|
       next if message[:kind] == :ignored && !PuppetLint.configuration.show_ignored
 
       message[:KIND] = message[:kind].to_s.upcase
-      message[:linenumber] = message[:line]
 
       if message[:kind] == :fixed || [message[:kind], :all].include?(configuration.error_level)
-        format_message message
-        print_context(message) if configuration.with_context
+        if configuration.json
+          message['context'] = get_context(message) if configuration.with_context
+          json << message
+        else
+          format_message(message)
+          print_context(message) if configuration.with_context
+        end
       end
     end
+    puts JSON.pretty_generate(json) if configuration.json
+
+    $stderr.puts 'Try running `puppet parser validate <file>`' if problems.any? { |p| p[:check] == :syntax }
   end
 
   # Public: Determine if PuppetLint found any errors in the manifest.
@@ -160,9 +184,7 @@ class PuppetLint
   # Returns nothing.
   # Raises PuppetLint::NoCodeError if no manifest code has been loaded.
   def run
-    if @code.nil?
-      raise PuppetLint::NoCodeError
-    end
+    raise PuppetLint::NoCodeError if @code.nil?
 
     if @code.empty?
       @problems = []
@@ -181,7 +203,7 @@ class PuppetLint
   #
   # Returns nothing.
   def print_problems
-    report @problems
+    report(@problems)
   end
 
   # Public: Define a new check.
